#include <fstream>
#include <span>
#include <stdexcept>

// a simple generic Matrix class
// templates in c++ allows genericity
// https://en.cppreference.com/w/cpp/language/templates
template <typename T>
class Matrix {
   public:
    // constructor, allocate linear array and
    // arrange pointer for 2D array
    Matrix<T>(int m, int n) {
        this->m = m;
        this->n = n;
        mat = new T *[m];
        mat[0] = new T[m * n];
        for (int i = 0; i < m; i++) {
            mat[i] = &(mat[0][i * n]);
        }
    }
    // accessor [] of class Matrix returns a pointer
    // to the corresponding line
    T *operator[](int index) {
        if (index < 0 || index >= m) {
            throw std::out_of_range("Index out of range");
        }
        return mat[index];
    }
    // destructor of class Matrix
    // automatically called when object goes out of scope
    ~Matrix() {
        delete[] mat[0];
        delete[] mat;
    }
    int cols() {
        return n;
    }
    int rows() {
        return m;
    }

    std::span<T> dataSpan() {
        return std::span<T>(this->mat[0], n * m);
    }

   private:
    int m, n;
    T **mat;
};

// initialise a Matrix
// here, & is the symbol for reference
// which allows passing an object by reference
// without using pointer syntax
// https://en.cppreference.com/w/cpp/language/reference
template <typename T>
void init_matrix(Matrix<T> &mat) {
    for (int i = 0; i < mat.rows(); i++) {
        for (int j = 0; j < mat.cols(); j++) {
            mat[i][j] = 0;
        }
    }
    for (int i = 0; i < mat.rows(); i++) {
        mat[i][0] = 0;
        mat[i][mat.cols() - 1] = 1;
    }
    for (int j = 0; j < mat.cols(); j++) {
        mat[0][j] = 1;
        mat[mat.rows() - 1][j] = 0;
    }
}

template <typename T>
void print_mat(Matrix<T> &m) {
    for (int i = 0; i < m.rows(); i++) {
        for (int j = 0; j < m.cols(); j++) {
            std::cout << m[i][j] << ", ";
        }
        std::cout << std::endl;
    }
}

template <typename T>
void write_mat(Matrix<T> &m, std::string filename) {
    std::ofstream file(filename);
    if (!file.is_open()) {
        std::cerr << "Failed to open file for writing: " << filename << std::endl;
        return;
    }
    for (int i = 0; i < m.rows(); i++) {
        for (int j = 0; j < m.cols(); j++) {
            file << m[i][j];
            if (j != m.cols() - 1) {
                file << ",";
            }
        }
        file << std::endl;
    }
    file.close();
}

template <typename T>
T sum_mat(Matrix<T> &m) {
    T sum = 0;
    for (int i = 0; i < m.rows(); i++) {
        for (int j = 0; j < m.cols(); j++) {
            sum += m[i][j];
        }
    }
    return sum;
}