#include <execution>
#include <iomanip>
#include <iostream>
#include <ranges>
#include <vector>

// Experimental CPP features
#include "includes/cartesian_product.hpp"
#include "includes/mdspan.hpp"

constexpr int dim_x = 8;
constexpr int dim_y = 8;

int main() {
    std::vector<int> data{dim_x * dim_y};

    auto md = std::mdspan{data.data(), dim_x, dim_y};

    // We create a view for x & y indexes
    auto xs = std::views::iota(0, dim_x);
    auto ys = std::views::iota(0, dim_y);

    // We build the indexes using the cartesian product
    auto idxs = std::views::cartesian_product(xs, ys);

    // We perform our operations on GPU
    std::for_each(std::execution::par_unseq,
                  idxs.begin(),
                  idxs.end(),
                  // Here we need to do a copy capture so that compiler knows that
                  // we will use the underlying memory on the GPU.
                  [md](auto tup) {
                      const auto& [x, y] = tup;
                      md(x, y) = x * y;
                  });

    // We print the result, since it's performed on CPU,
    // we can simply capture by reference.
    std::for_each(xs.begin(), xs.end(), [&md, &ys](auto i) {
        std::for_each(ys.begin(), ys.end(), [&md, i](auto j) {
            std::cout << std::setw(3) << md(i, j);
        });
        std::cout << "\n";
    });

    std::cout << "\n";
}
