// https://polylux.dev/book/polylux.html
// https://typst.app/docs
#import "@preview/polylux:0.3.1": *
#import themes.metropolis : *

#show: metropolis-theme.with(aspect-ratio: "16-9", footer: "Parallel STL - HPC 2024")

#set text(font: "Fira Sans", weight: "light", size: 20pt)
#show math.equation: set text(font: "Fira Math")

#title-slide(
  subtitle: "Algorithmes parallèles STL avec C++ 20",
  title: "GPGPU",
  author: "Michaël El Kharroubi",
  date: "27.02.2024",
  extra: "HPC 2024 - HEPIA",
)

#new-section-slide("Introduction")

#slide(
  title: "Définition",
)[
  Le terme "_general-purpose computing on graphics processing units_" (GPGPU)
  désigne l'ensemble des programmmes qui ne sont pas directement liés au calcul
  graphique et qu'on réalise sur GPU.

  Par exemple :
  - Simulation de phénomènes physiques
  - IA et Machine learning
  - Minage de crypto-monnaie
]

#slide(
  title: "Pourquoi a-t-on besoin de GPU?",
)[
  Un CPU est conçu pour suivre une série d'instructions séquentiellement (ce qui
  n'est plus le cas, dans une moindre mesure, avec les instructions vectorielles
  ex: AVX).

  #box(columns(2, gutter: 22pt)[
    == Adapté :
    - Ordonnancer des tâches
    - Accéder aux périphériques
      - Réseau
      - Disque
      - Clavier/souris
    #colbreak()
    == Moins adapté :
    - Calcul vectoriel
    - Calcul matriciel
  ])

  Les GPU apportent une réponse à ces problèmes
]

#slide(
  title: "CPU vs GPU",
)[
  Un GPU est un type de processeur conçu pour effectuer massivement des calculs en
  parallèle

  #figure(
    image("assets/cpu-gpu.svg", width: 70%),
    caption: [CPU vs GPU Architecture - Source : NVIDIA CUDA Programming Guide version 3.0 -
      2010],
  )

  Voici une petite vidéo du partenariat entre Myth Busters et Nvidia, pour
  visualiser la différence #link("https://youtu.be/-P28LKWTzrI")
]

#new-section-slide("La programmation sur GPU")

#slide(
  title: "CUDA",
)[
  CUDA est le kit de développement de NVIDIA pour faire du GPGPU. C'est le kit le
  plus populaire de développement GPGPU.

  CUDA permet d'écrire du code dans un dialecte C/C++ et de le compiler pour qu'il
  soit exécuté sur des GPU de Nvidia.

  Dans la concurrence, on peut également citer OpenCL (multi-plateforme), ROCm
  pour AMD ou OneAPI pour Intel.
]

#slide(title: "Déroulement classique d'un programmme GPGPU")[
  + Le CPU envoie les données à traiter au GPU #pause
  + Le GPU applique un kernel sur les données #pause
  + Le CPU récupère le résultat
]

#new-section-slide("Exemple de code CUDA et notions à retenir")

#slide(
  title: "Notions essentielles à retenir",
)[
  - Le CPU et le GPU sont deux périphériques distincts
  - Le GPU est prévu pour effectuer massivement des calculs
  - La mémoire n'est pas partagée, l'espace d'adressage n'est donc pas commun
  - On parle de *Host* pour le CPU et de *Device* pour le GPU
]

#new-section-slide("Questions ?")